/*****************************************************************************
 * \file mi32/bintoreg.h
 * \brief A Binary raster to region converer class
 *
 * \if NODOC
 * $Log: bintoreg.h_v $
 * Revision 1.7  2012/12/05 14:13:31  mju
 * Change v_processPolygon to use polyline and add xpoints array.
 *
 * Revision 1.6  2011/10/21 20:48:04  mju
 * Remove redundant hdr.
 *
 * Revision 1.5  2006/09/20 17:28:18  scowan
 * Did not need polyline include file.
 *
 * Revision 1.4  2005/03/09 22:56:20  mju
 * Eliminate most virtual overrides, now all goes through ProcessLine.
 *
 * Revision 1.3  2004/09/24 15:01:17  mju
 * Add isPointOnSpan.
 *
 * Revision 1.2  2004/07/02 17:38:29  dwilliss
 * Added AddSpansFromBytes
 *
 * Revision 1.1  2003/12/15 17:26:58  dwilliss
 * Initial revision
 * \endif
**/

#ifndef INC_MI32_BINTOREG_H
#define INC_MI32_BINTOREG_H

#ifndef INC_MI32_REGION2D_H
#include <mi32/region2d.h>
#endif


//! Class for converting a binary raster to a region.
//!
//! This is a generic base class which can be extended to create any sort of
//! region: REGION2D, X region or Windows Region.
class BINARYTOREGIONINTERFACE {
	public:

		//! Add a span to the region.
		//! This is used to transfer the binary object (whatever the source)
		//! into a format that this class can use.
		//! Note, this method makes sure that the spans are kept in order and
		//! combines spans which overlap. 
		//! If you know you're adding spans in top-to-bottom, left-to-right order,
		//! call AddSpanFast() instead.
		ERRVALUE AddSpan (
			INT32 x1,
			INT32 x2,
			INT32 y
			);

		//! Add a span to the region.
		//! This is used to transfer the binary object (whatever the source)
		//! into a format that this class can use.
		//! Note, no attempt is made to keep the spans sorted or make sure the 
		//! span does not overlap any existing spans.  If you know you're adding
		//! spans in top-to-bottom, left-to-right order, you can use this method
		//! and it will be much faster than AddSpan()
		ERRVALUE AddSpanFast (
			INT32 x1,
			INT32 x2,
			INT32 y
			);

		//! Add spans to the region from a buffer of UINT8s
		//!
		//! This function takes an array of UINT8s representing a raster line
		//! and adds spans for all of the non-zero runs.
		//! The buffer is treated as one byte per pixel with 0 being false and
		//! non-zero being true.
		//! 
		//! The bFast parameter tells the function whether or not you're procssing
		//! lines in a top-down order without back-tracking.  If you are, pass
		//! true and it can add the spans without having to keep them sorted.
		ERRVALUE AddSpansFromBytes (
			INT32 lin,
			INT32 col,
			INT32 len,
			const UINT8 *buf,
			bool bFast = false
			);

		//! Clear any existing spans.
		void Clear ();

		//! Determine if specified point is on existing span.
		bool IsPointOnSpan (
			INT32 x,
			INT32 y
			) const;

		//! Process the spans that have been set.
		ERRVALUE ProcessSpans ();

	protected:

		BINARYTOREGIONINTERFACE ();

		virtual ~BINARYTOREGIONINTERFACE ();

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		typedef BINARYTOREGIONINTERFACE THISCLASS;

		struct SPAN {
			INT32 start;
			INT32 end;
			};

		typedef SIMPLE_ARRAY<SPAN> SPANARRAY;

		SIMPLE_ARRAY<SPANARRAY*> m_Spans;

		ERRVALUE FillSpan (INT32 x1, INT32 x2, INT32 y);
		int TraceTest (INT32 x, INT32 y);

		#endif	// GENERATING_DOXYGEN_OUTPUT

		// Overridables

		//! Clear the current region.
		//! Will be called by the Clear() method, which also gets called by
		//! the destructor.
		virtual void v_Clear ();

		//! Add or remove a line from the region.
		//! Note that the points array passed in is not const for a reason. 
		//! The method is allowed to modify the points in the buffer before 
		//! adding them to the region.  Just don't free the buffer.
		virtual ERRVALUE v_ProcessPolygon (
			const POLYLINE& polyline
			) = 0;

	};


// An implementation of BINARYTOREGIONINTERFACE which converts to a REGION2D.
class BINARYTOREGION2D : public BINARYTOREGIONINTERFACE {
	public:
		BINARYTOREGION2D() {}
		~BINARYTOREGION2D() {}

		// Get the region.
		// You must setup the spans and call ProcessSpans first.
		const REGION2D& GetRegion () const
			{ return (m_region); }

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		REGION2D m_region;

		void v_Clear();
		virtual ERRVALUE v_ProcessPolygon (const POLYLINE& polyline);
		#endif
	};


#ifdef X_NATIVE

#include <X11/Intrinsic.h>

// An implementation of BINARYTOREGIONINTERFACE which converts to an X Region.
class BINARYTOXREGION : public BINARYTOREGIONINTERFACE {
	public:
		BINARYTOXREGION() :
			m_region(0)
			{}

		~BINARYTOXREGION() {}

		Region GetRegion () const
			{ return (m_region); }

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		Region m_region;
		SIMPLE_ARRAY<XPoint> m_xpoints;

		void v_Clear();
		virtual ERRVALUE v_ProcessPolygon (const POLYLINE& polyline);
		#endif	// GENERATING_DOXYGEN_OUTPUT
	};


#endif

#endif // INC_MI32_BINTOREG_H
