/**
 * \file <mi32/colorspread.h>
 * \brief COLORSPREAD definitions.
 *
 * \if NODOC
 * $Id: colorspread.h_v 1.1 2010/09/02 17:36:59 mju Exp $
 *
 * $Log: colorspread.h_v $
 * Revision 1.1  2010/09/02 17:36:59  mju
 * Initial revision
 *
 * \endif
**/

#ifndef	INC_MI32_COLORSPREAD_H
#define	INC_MI32_COLORSPREAD_H

#ifndef	INC_MI32_MISTRING_H
#include <mi32/mistring.h>
#endif

#ifndef	INC_MI32_COLOR_H
#include <mi32/color.h>
#endif

#ifndef	GENERATING_DOXYGEN_OUTPUT
class COLOR_ARRAY;
class SERIALIZERITEM;
#endif

//! Color spread definition.
class COLORSPREAD {
	public:

		//! Spread type
		enum TYPE {
			TYPE_Manual =		0,
			TYPE_Standard =	1,
			TYPE_User =			2
			};

		enum PATH {
			PATH_Longest =			0,
			PATH_RGB,
			PATH_HISCW,
			PATH_HISCCW,
			PATH_HBSCW,
			PATH_HBSCCW,
			PATH_COUNT
			};

		//! Get serializer item definition array.
		static const SERIALIZERITEM* SerialGetItemDef ();

		//! Default constructor.
		COLORSPREAD ();

		//! Destructor.
		~COLORSPREAD ();

		//! Compute color palette.
		ERRVALUE ComputePalette (
			COLOR_ARRAY& palette,		//!< Palette to contain result
			int NumColors					//!< Number of colors in resulting array, 0 to use as originally defined for TYPE_Standard and TYPE_User
			) const;

		//! Get beginning color for manual spread.
		const COLOR& GetColorBegin () const
			{ return (m_ColorBegin); }

		//! Get ending color for manual spread.
		const COLOR& GetColorEnd () const
			{ return (m_ColorEnd); }

		//! Get name for standar/user spread.
		const MISTRING& GetName () const
			{ return (m_Name); }

		//! Get path for manual spread.
		PATH GetPath () const
			{ return (m_Path); }

		//! Get spread type.
		TYPE GetType () const
			{ return (m_Type); }

		//! Read values from INI file.
		void IniRead (
			INIHANDLE *inihnd,
			const char *IniGroup,
			const char *IniName
			);

		//! Write values to INI file.
		void IniWrite (
			INIHANDLE *inihnd,
			const char *IniGroup,
			const char *IniName
			) const;

		//! Get whether reversed.
		bool IsReversed () const
			{ return (m_IsReversed); }

		//! Set beginning color for manual spread.
		void SetColorBegin (
			const COLOR& color
			) { m_ColorBegin = color; }

		//! Set ending color for manual spread.
		void SetColorEnd (
			const COLOR& color
			) { m_ColorEnd = color; }

		//! Set to manual spread with start/end colors and path.
		void SetManual (
			const COLOR& ColorBegin,
			const COLOR& ColorEnd,
			PATH path
			) { m_Type = TYPE_Manual; m_ColorBegin = ColorBegin; m_ColorEnd = ColorEnd; m_Path = path; }

		//! Set as "named" spread with TYPE_Standard or TYPE_User.
		void SetNamed (
			const MISTRING& name,
			TYPE type = TYPE_Standard
			) { m_Type = type; m_Name = name; }

		//! Set manual path.
		void SetPath (
			PATH path
			) { m_Path = path; }

		//! Set whether reversed if not manual.
		void SetReversed (
			bool reversed					//!< True if reversed, False if not
			) { m_IsReversed = reversed; }

		//! Set type.
		void SetType (
			TYPE type						//!< Spread type
			) { m_Type = type; }

	private:
		#ifndef	GENERATING_DOXYGEN_OUTPUT
		TYPE m_Type;
		PATH m_Path;
		COLOR m_ColorBegin;
		COLOR m_ColorEnd;
		MISTRING m_Name;
		bool m_IsReversed;
		#endif
	};

#ifndef GENERATING_DOXYGEN_OUTPUT
DEFINE_ENUM_OP_INCREMENT(COLORSPREAD::PATH)
#endif

#endif	// INC_MI32_COLORSPREAD_H

