/**
 * \file mi32/lidarclassifier.h
 * \brief LIDAR::CLASSIFIER definitions
 *
 * \if NODOC
 * $Id: lidarclassifier.h_v 1.7 2012/09/28 21:24:36 mju Exp $
 *
 * $Log: lidarclassifier.h_v $
 * Revision 1.7  2012/09/28 21:24:36  mju
 * Change pointbuffer to keep status for points and only set new class when apply.
 *
 * Revision 1.6  2012/09/27 15:38:45  mju
 * Add get/setProcessWithheld.
 *
 * Revision 1.5  2012/09/24 20:24:01  mju
 * Add control of whether can choose input classes.
 *
 * Revision 1.4  2012/09/21 15:12:15  mju
 * Change how search grid finds neighbors to use nearest first.
 *
 * Revision 1.3  2012/08/28 13:36:09  mju
 * Add m_duplicateSet.
 *
 * Revision 1.2  2012/08/27 19:53:51  mju
 * Change jobGetValues/Read to not be inline.
 *
 * Revision 1.1  2012/08/24 21:25:26  mju
 * Initial revision
 *
 * \endif
**/

#ifndef	INC_MI32_LIDARCLASSIFIER_H
#define	INC_MI32_LIDARCLASSIFIER_H

#ifndef	INC_MI32_LIDARCLASSDESC_H
#include <mi32/lidarclassdesc.h>
#endif

#ifndef	INC_MI32_DOUBLEAR_H
#include <mi32/doublear.h>
#endif

#ifndef	INC_MI32_SIMPLEAR_H
#include <mi32/simplear.h>
#endif

#ifndef	INC_MI32_MILIST_H
#include <mi32/milist.h>
#endif

#ifndef	INC_MI32_UNSIGNEDSET_H
#include <mi32/unsignedset.h>
#endif

#ifndef	INC_MI32_RECT_H
#include <mi32/rect.h>
#endif

#ifndef	INC_MAP
#define	INC_MAP
#include <map>
#endif


// Forward declarations.
#ifndef GENERATING_DOXYGEN_OUTPUT
class MIJOB;
namespace LIDAR {
	namespace LAS {
		class DATASET;
		class POINTRECORD;
		}
	}
#endif


namespace LIDAR {

//=====================================================================================================================

//! Base classifier class.
class CLASSIFIER {
	public:

		// Forward declarations.
		class CLASSRESULTS;
		class CLASSRESULTSLIST;
		class CLASSMAP;
		class POINTBUFFER;
		class SEARCHGRID;

		//! Destructor.
		virtual ~CLASSIFIER ();

		//! Access list of inputs.
		CLASSRESULTSLIST& AccessInputs ()
			{ return (*m_pInputs); }

		//! Add input class.
		void AddInput (
			const CLASSID& classid,					//!< Class identifier
			const CLASSDESCSET& ClassDescSet
			);

		//! Get whether user is allowed to choose which input classes to process.
		bool GetAllowChooseInputClasses () const
			{ return (m_AllowChooseInputClasses); }

		//! Get list of inputs.
		const CLASSRESULTSLIST& GetInputs () const
			{ return (*m_pInputs); }

		//! Get margin to add around extent for selecting reference points.
		double GetMargin () const
			{ return (v_GetMargin()); }

		//! Get maximum points desired to process.
		UINT64 GetMaxPointsDesired () const
			{ return (v_GetMaxPointsDesired()); }

		//! Get list of outputs.
		const CLASSRESULTSLIST& GetOutputs () const
			{ return (*m_pOutputs); }

		//! Get whether to process "withheld" points.
		bool GetProcessWithheld () const
			{ return (m_ProcessWithheld); }

		//! Read all settings from Ini file.
		void IniRead (
			INIHANDLE inihnd = 0
			);

		//! Write all settings to Ini file.
		void IniWrite (
			INIHANDLE inihnd = 0
			) const;

		//! Get values into MIJOB for job processing.
		ERRVALUE JobGetValues (
			MIJOB& mijob
			) const;

		//! Read values from MIJOB for job processing.
		ERRVALUE JobRead (
			const MIJOB& mijob
			);

		//! Reset all parameters and input classes.
		void Reset ();

		//! Setup result class numbers for all inputs based on current usage and CLASSDESCSET.
		void SetupClassNumbersAll (
			const CLASSDESCSET& ClassDescSet
			);

		//! Set whether to process "withheld" points.
		//! Default is to not process.
		void SetProcessWithheld (
			bool ProcessWithheld
			) { m_ProcessWithheld = ProcessWithheld; }

		//! Setup result class numbers for specified input based on current usage and CLASSDESCSET.
		void SetupClassNumbersInput (
			const CLASSDESCSET& ClassDescSet,
			CLASSRESULTS& Input
			);

	protected:

		//! Constructor.
		CLASSIFIER (
			const MISTRING& name,
			const MISTRING& info,
			const char *inigroup
			);

		//! Add default input class.
		//! All outputs classes must be added before any input classes.
		void AddDftInput (
			CLASSDESC::USAGE usage,							//!< Input class usage
			CLASSDESC::USAGE DftResult1 = CLASSDESC::USAGE_Unspecified,	//!< Default result override for first output class
			CLASSDESC::USAGE DftResult2 = CLASSDESC::USAGE_Unspecified,	//!< Default result override for second output class
			CLASSDESC::USAGE DftResult3 = CLASSDESC::USAGE_Unspecified,	//!< Default result override for third output class
			CLASSDESC::USAGE DftResult4 = CLASSDESC::USAGE_Unspecified,	//!< Default result override for fourth output class
			CLASSDESC::USAGE DftResult5 = CLASSDESC::USAGE_Unspecified	//!< Default result override for fifth output class
			);

		//! Add output class with allowed results.
		//! All outputs classes must be added before any input classes.
		void AddOutput (
			CLASSDESC::USAGE usage,							//!< Output class usage
			bool AllowLeaveUnchanged = true,				//!< Allow option to leave unchanged
			CLASSDESC::USAGE ResultOther1 = CLASSDESC::USAGE_Unspecified,
			CLASSDESC::USAGE ResultOther2 = CLASSDESC::USAGE_Unspecified
			);

		//! Set whether user is allowed to choose which input classes to process.
		void SetAllowChooseInputClasses (
			bool AllowChooseInputClasses
			) { m_AllowChooseInputClasses = AllowChooseInputClasses; }

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		MISTRING m_Name;
		MISTRING m_Description;
		const char *m_IniGroup;
		CLASSRESULTSLIST *m_pDftInputs;
		CLASSRESULTSLIST *m_pInputs;
		CLASSRESULTSLIST *m_pOutputs;
		bool m_AllowChooseInputClasses;
		bool m_ProcessWithheld;
		#endif

		//! Get margin for processing by block.
		virtual double v_GetMargin () const;

		//! Get maximum desirable number of points to process.
		virtual UINT64 v_GetMaxPointsDesired () const = 0;

		//! Read all settings from Ini file.
		virtual void v_IniRead (
			INIHANDLE inihnd
			);

		//! Write all settings to Ini file.
		virtual void v_IniWrite (
			INIHANDLE inihnd
			) const;

		//! Get values into MIJOB for saving.
		virtual void v_JobGetValues (
			MIJOB& mijob
			) const = 0;

		//! Read setting values from MIJOB.
		virtual ERRVALUE v_JobRead (
			const MIJOB& mijob
			) = 0;

		//! Reset input parameters.
		virtual void v_Reset ();

	};

//=====================================================================================================================

//! Container for class "usage", array of "results" and indication of whether is "used".
class CLASSIFIER::CLASSRESULTS {
	public:

		//! Default constructor.
		CLASSRESULTS ();

		//! Construct with CLASSID.
		explicit CLASSRESULTS (
			const CLASSID& classid
			);

		//! Copy constructor.
		CLASSRESULTS (
			const CLASSRESULTS& rhs
			);

		//! Destructor.
		~CLASSRESULTS ();

		//! Assignment.
		CLASSRESULTS& operator= (
			const CLASSRESULTS& rhs
			);

		//! Add result.
		void AddResult (
			CLASSDESC::USAGE result			//!< Result usage
			) { m_Results.Append(CLASSID(result)); }

		//! Get class identifier.
		const CLASSID& GetClass () const
			{ return (m_classid); }

		//! Get result for specified index.
		const CLASSID& GetResult (
			UINT16 idx
			) const { return (m_Results[idx]); }

		//! Get array of results.
		const SIMPLE_ARRAY<CLASSID>& GetResults () const
			{ return (m_Results); }

		//! Get indication of whether is "used".
		bool IsUsed () const
			{ return (m_IsUsed); }

		//! Set input class number.
		void SetClassNumber (
			UINT16 ClassNumber
			) { m_classid.SetNumber(ClassNumber); }

		//! Set result by index.
		void SetResult (
			UINT16 idx,								//!< Index in result array
			const CLASSID& result				//!< Result class identifier
			) { m_Results[idx] = result; }

		//! Set whether is "used".
		void SetUsed (
			bool IsUsed
			) { m_IsUsed = IsUsed; }

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		SIMPLE_ARRAY<CLASSID> m_Results;
		CLASSID m_classid;
		bool m_IsUsed;
		#endif
	};

//=====================================================================================================================

//! List of CLASSRESULTS.
class CLASSIFIER::CLASSRESULTSLIST : public MILIST<CLASSRESULTS> {
	public:

		//! Constructor.
		CLASSRESULTSLIST ();

		//! Destructor.
		~CLASSRESULTSLIST ();

		//! Find item with specified usage (const).
		CONST_ITERATOR FindItem (
			CLASSDESC::USAGE usage
			) const;

		//! Find item with specified usage (non-const).
		ITERATOR FindItem (
			CLASSDESC::USAGE usage
			);

		//! Find item with specified class number (const).
		CONST_ITERATOR FindItem (
			UINT16 ClassNumber			//!< Class number
			) const;

		//! Find item with specified class number (non-const).
		ITERATOR FindItem (
			UINT16 ClassNumber			//!< Class number
			);

		//! Read list from Ini group.
		void IniRead (
			INIHANDLE inihnd,				//!< Ini file handle
			const char *inigroup,		//!< Group, typically unique to classifier implementation
			const char *inibasename,	//!< Ini base name for list entry
			INT32 NumResults				//!< Number of "result" values to use
			);

		//! Write list to Ini group.
		void IniWrite (
			INIHANDLE inihnd,				//!< Ini file handle
			const char *inigroup,		//!< Group, typically unique to classifier implementation
			const char *inibasename		//!< Ini base name for list entry
			) const;

	};

//=====================================================================================================================

class CLASSIFIER::CLASSMAP {
	public:

		CLASSMAP ();

		~CLASSMAP ();

		UINT8 operator() (
			UINT32 output,
			UINT8 ClassNumInput
			) const { return (m_Outputs[output][ClassNumInput]); }

		UINT8 GetNumOutputs () const
			{ return (m_NumOutputs); }

		ERRVALUE Initialize (
			const CLASSRESULTSLIST& InputList,
			UINT16 NumOutputs
			);

		bool IsInputUsed (
			UINT8 ClassNumInput
			) const { return (m_IsInputUsed[ClassNumInput]); }

		bool IsPointRecordUsedInput (
			const LAS::POINTRECORD& pointRec
			) const;

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		SIMPLE_ARRAY<UINT8> *m_Outputs;
		bool m_IsInputUsed[256];
		UINT8 m_NumOutputs;
		#endif
	};
	
//=====================================================================================================================

class CLASSIFIER::POINTBUFFER {
	public:

		enum STATUS {
			STATUS_Duplicate =	254,
			STATUS_Unchanged =	255
			};

		//! Constructor.
		POINTBUFFER ();

		//! Destructor.
		~POINTBUFFER ();

		//! Begin adding points.
		ERRVALUE AddBegin (
			UINT32 NumReserve					//!< Initial number of points to reserve
			);

		//! Finish adding points.
		//! Merges margin and main points/classes and sets status for all points to STATUS_Unchanged.
		ERRVALUE AddFinish ();

		//! Add point within extent to be classified.
		//! Points should be added in ascending order by PointNum.
		ERRVALUE AddPoint (
			UINT64 PointNum,					//!< Point number in dataset
			const DPOINT3D& pt,				//!< Point coordinates
			UINT8 ClassOld						//!< Current classification
			);

		//! Add point in margin area, index not retained so won't be altered by ApplyClassification.
		ERRVALUE AddPointToMargin (
			const DPOINT3D& pt,				//!< Point coordinates
			UINT8 ClassOld						//!< Current classification
			);

		//! Apply classification to all changed points based on specified map.
		ERRVALUE ApplyClassification (
			LIDAR::LAS::DATASET& dataset,	//!< Dataset
			const CLASSMAP& ClassMap		//!< Classification map
			);

		//! Calculate index of points sorted by X,Y,Z
		ERRVALUE CalcIndexSortedXYZ (
			SIMPLE_ARRAY<UINT32>& PointIdx	//!< Index array returned
			) const;

		//! Get extents of all points.
		const DRECT3D& GetExtents () const
			{ return (m_Extents); }

		//! Get number of points in buffer.
		UINT32 GetNumPoints () const
			{ return (m_Points.GetNumItems()); }

		//! Get point coordinates.
		const DPOINT3D& GetPoint (
			UINT32 idx							//!< Point index within buffer
			) const { return (m_Points[idx]); }

		//! Get set of point numbers as added via AddPoint.
		const UNSIGNEDSET& GetPointNumSet () const
			{ return (m_PointNumSet); }

		//! Get output value or status for index.
		//! @return STATUS_Unchanged, STATUS_Duplicate or "output" value.
		UINT8 GetOutputStatus (
			UINT32 idx
			) const { return (m_Status[idx]); }

		//! Set point as duplicate of another point.
		//! Duplicate point will have STATUS_Duplicate.
		void SetDuplicate (
			UINT32 idxOriginal,				//!< Index within buffer of original point
			UINT32 idxDuplicate				//!< Index within buffer of duplicate point
			);

		//! Set output value for point
		void SetOutput (
			UINT32 idx,							//!< Point index within buffer
			UINT8 output						//!< Output value
			) { m_Status[idx] = output; }

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		typedef std::multimap<UINT32,UINT32> MAP_DUPLICATES;

		DOUBLE_ARRAY<DPOINT3D> m_Points;
		DOUBLE_ARRAY<DPOINT3D> m_PointsMargin;
		SIMPLE_ARRAY<UINT8> m_ClassOld;
		SIMPLE_ARRAY<UINT8> m_ClassOldMargin;
		SIMPLE_ARRAY<UINT8> m_Status;
		UNSIGNEDSET m_PointNumSet;
		MAP_DUPLICATES m_MapDuplicates;
		BITSET m_DuplicateSet;
		DRECT3D m_Extents;

		static int SortComparePointXYZ (UINT32 *pIdx1, UINT32 *pIdx2, void *vpPointBuffer);

		#endif
	};

//=====================================================================================================================

class CLASSIFIER::SEARCHGRID {
	public:

		class RESULTSET : public UNSIGNEDSET {
			public:

				RESULTSET ();

				~RESULTSET ();

				//! Add next neighbor point.
				//! @return Index (in point buffer) of point added or error < 0.
				INT32 AddNeighborPoint ()
					{ return (AddNeighborPoints(1)); }

				//! @return Index (in point buffer) of last point added or error < 0.
				INT32 AddNeighborPoints (
					UINT32 nPointsAdd
					);

				void Initialize (
					const SEARCHGRID *pGrid,
					UINT32 gridrow,
					UINT32 gridcol
					);

			private:
				#ifndef GENERATING_DOXYGEN_OUTPUT
				const SEARCHGRID *m_pGrid;
				UINT32 m_row;
				UINT32 m_col;
				UINT32 m_neighborhoodSize;
				UINT32 m_idxNeighbor;
				DPOINT2D m_center;
				SIMPLE_ARRAY<UINT32> m_NeighborPointIdxs;
				SIMPLE_ARRAY<UINT32> m_NeighborSortIdxs;
				DOUBLE_ARRAY<double> m_NeighborDistances;

				static int SortCompare (UINT32 *pSortIdx1, UINT32 *pSortIdx2, void*);
				#endif
			};

		//! Constructor.
		SEARCHGRID ();

		//! Destructor.
		~SEARCHGRID ();

		void CalcCellCenter (
			UINT32 row,
			UINT32 col,
			DPOINT2D& center
			) const {
			center.x = m_origin.x + (col + .5) * m_cellsize.x;
			center.y = m_origin.y + (row + .5) * m_cellsize.y;
			}

		//! Get (const) set of point indices for specified grid cell.
		const UNSIGNEDSET& GetCellSet (
			UINT32 row,
			UINT32 col
			) const { return (m_CellSet[row*m_ncols+col]); }

		const DPOINT2D& GetCellSize () const
			{ return (m_cellsize); }

		UINT32 GetNumCols () const
			{ return (m_ncols); }

		UINT32 GetNumRows () const
			{ return (m_nrows); }

		const DPOINT2D& GetOrigin () const
			{ return (m_origin); }

		ERRVALUE GetResultSet (
			RESULTSET& resultset,		//!< Result set returned
			UINT32 row,						//!< Row in grid
			UINT32 col,						//!< Column in grid
			UINT32 minpoints = 0			//!< Minimum number of points desired
			) const;

		ERRVALUE Initialize (
			const DPOINT2D& origin,		//!< Grid origin in point coordinates
			const DPOINT2D& cellsize,	//!< Grid cell size in point coordinate space
			UINT32 nrows,					//!< Number of grid rows
			UINT32 ncols,					//!< Number of grid columns
			bool IsOverlapped = false	//!< Whether to overlap grid cells (by 1/2 cell in each dimension)
			);

		bool IsOverlapped () const
			{ return (m_IsOverlapped); }

		//! Sort points into grid cells.
		ERRVALUE SortPointsToCells (
			const POINTBUFFER& PointBuffer,			//!< Source point buffer
			const SIMPLE_ARRAY<UINT32>& PointIdx	//!< Point indices to store
			);

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		UNSIGNEDSET *m_CellSet;
		const POINTBUFFER *m_pPointBuffer;
		DPOINT2D m_origin;
		DPOINT2D m_cellsize;
		UINT32 m_nrows;
		UINT32 m_ncols;
		bool m_IsOverlapped;

		UNSIGNEDSET& AccessCellSet (
			UINT32 row,
			UINT32 col
			) { return (m_CellSet[row*m_ncols+col]); }

		#endif
	};

//=====================================================================================================================

}	// End namespace LIDAR

#endif	// INC_MI32_LIDARCLASSIFIER_H
