/**
 * \file mi32/mithread.h
 * \brief MITHREAD class definitions
 *
 * \if NODOC
 * $Id: mithread.h_v 1.19 2012/08/24 20:18:10 mju Exp $
 *
 * $Log: mithread.h_v $
 * Revision 1.19  2012/08/24 20:18:10  mju
 * Add pool::suspendQueuedJobs.
 *
 * Revision 1.18  2011/11/02 21:19:08  mju
 * Remove redundant hdr.
 *
 * Revision 1.16  2008/06/16 20:47:04  mju
 * Add notification that job cancelled before started.
 *
 * Revision 1.15  2008/05/22 16:48:38  mju
 * Add worker getPriv for internal use.
 *
 * Revision 1.14  2008/05/16 16:21:53  mju
 * Change flag to be StopIfError so default is to continue.
 *
 * Revision 1.13  2008/03/20 16:04:50  mju
 * Add id parm to delegate for job/queue completion.
 *
 * Revision 1.12  2007/10/18 22:43:34  scowan
 * Added mi status manager references to worker, pool, and job ctors.
 * Setup status managers for threads specified by workers or jobs.
 *
 * Revision 1.9  2007/09/07 17:48:09  scowan
 * MOre fixes and added logical core use in pool class.
 *
 * Revision 1.8  2007/08/31 22:53:07  scowan
 * Added methods to start and wait for jobs.
 *
 * Revision 1.1  2005/05/17 16:42:08  scowan
 * Initial revision
 *
 * \endif
**/

#ifndef	INC_MI32_MITHREAD_H
#define	INC_MI32_MITHREAD_H

#ifndef	INC_MI32_DELEGATE_H
	#include <mi32/delegate.h>
#endif

#ifndef	INC_MI32_ERRHANDLER_H
	#include <mi32/errhandler.h>		// Needed for ERRORSTATE define
#endif

#ifdef MISYSTEMDLL
	#define CLASSLIBEXPORT MI_DLLCLASSEXPORT
#else
	#define CLASSLIBEXPORT MI_DLLCLASSIMPORT
#endif

#if defined(WIN32) && defined(AddJob)
#undef AddJob
#endif


#ifndef GENERATING_DOXYGEN_OUTPUT
// Forward declarations.
class MISTATUSMANAGER;
namespace MITHREAD {
	class JOB;
	class QUEUE;
	class POOL;
	struct POOLITEM;
	class POOLJOB;
	struct QUEUEITEM;
	struct QUEUEITEMLIST;
	}
#endif


//! Multithreading support.
namespace MITHREAD {

typedef UINT32 ID;

//! Delegate accepting MITHREAD::ID parameter.
typedef fastdelegate::FastDelegate<void(MITHREAD::ID)> DELEGATE_VOID_ID;

//! Thread class to manage a core level threading system.  Allows execution of a job.
class CLASSLIBEXPORT WORKER {
	public:
		//! Thread ctor, the thread instance will be created and put into a wait state
		explicit WORKER (
			MISTATUSMANAGER* DftManager = 0,		//!< Default status manager to assign any job, unless job overrides it
			const char* DebugName = 0				//!< Name to assign to the thread, only applicable to Windows debugging under Visual Studio.
			);

		//! Thread dtor
		//! This call will block if a job is running.
		~WORKER (
			);

		//! Cancel a running job in the thread.  The derived class of JOB must call JOB::CheckCancel() to determine if it has been canceled.			
		void CancelJob (
			);		

		//! Execute the job passed in in the thread instance managed by this class
		//! If the JOB::STATE_DeleteWhenCompleted is set on 'Job', the job will be deleted when completed.
		//! This call will block if another job is running.
		void ExecuteJob (
			JOB& Job
			);

		//! Determine if the currently running job is in the cancel state
		//! @return 'True' if the job is in the process of canceling, 'false' if no job is running or CancelJob() has not been called
		bool IsCanceling (
			) const;

	#ifndef GENERATING_DOXYGEN_OUTPUT
		class PRIV;
		PRIV* GetPriv () { return (m_pPriv); }
	#endif

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		WORKER (const WORKER&);
		WORKER& operator=(const WORKER&);
	
		PRIV* m_pPriv;
		friend class JOB;
		friend class POOL;
		#endif // GENERATING_DOXYGEN_OUTPUT
	};


//! Class to manage multiple threads in a 'pool'.  The list of jobs to do, called 'QUEUE', is processed to execute the jobs 
//! in an available active thread in the pool.  After a JOB is completed, it is removed from its queue.  When a queue is empty, 
//! it is removed from its parent queue.
class CLASSLIBEXPORT POOL {
	public:

		//! Ctor
		POOL (
			UINT32 NumPoolThreads = 0,				//!< The number of pool worker threads to allocate.  Default is the number of processor cores.
			MISTATUSMANAGER* DftManager = 0,		//!< Default status manager to assign any job, unless job overrides it
			bool IncludeLogicalCores = false
			);
		
		//! Dtor
		//! This call will block if jobs are running.
		~POOL (
			);
			
		//! Add a queue to the pool.
		//! @return The ID of the queue added
		ID AddQueue (
			QUEUE& Queue,
			bool AddToFront = false			//!< If 'true', places the job queue ahead of all other queues and executes it ASAP.
			);

		//! Add a queue to the pool in front of the Job given.
		//! @return The ID of the queue added
		ID AddQueue (
			QUEUE& Queue,
			JOB& Job									//!< If 'Job' adds a QUEUE, this must be set to 'Job' so that the 'Queue' can be properly 
														//!< placed in front of 'Job' and 'Job' blocked until its queue is done.  NULL refers to the 
														//!< thread that owns 'this' POOL.
			);

		//! Cancel the job referenced by 'JobID'.
		//! Will immediately cancel a job that is waiting to run.  If a job is running, will call WORKER::CancelJob()
		void CancelJob (
			ID JobID							//!< ID of the job
			);
			
		//! Cancel all jobs and queues referenced by 'QueueID'.
		//! Will immediately cancel jobs that are waiting to run.  If jobs are running, will call WORKER::CancelJob() for each job.
		void CancelQueue (
			ID QueueID							//!< ID of the queue
			);
			
		//! Get the number of active threads in the pool.  "Active Threads" are threads that are currently or can be executing MITHREAD::JOB's.
		//! If a thread is created because a JOB is blocked waiting for other jobs, that thread is not counted as active.
		//! @return Number of active threads in the pool.
		UINT32 GetNumActiveThreads () const;
	
		//! Set a delegate that gets called when all the jobs are complete.
		void SetDelegateOnCompletion (
			DELEGATE_VOID_NOPARMS OnCompletion
			);
			
		//! Start the jobs added via the AddQueue() methods.
		void StartQueuedJobs ();

		//! Suspend jobs that have not already started.
		//! Jobs will remain in the queue(s)
		void SuspendQueuedJobs ();
			
		//! Wait until all of the jobs in the queue are completed
		void WaitForCompletion ();
			
	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		POOL (const POOL&);
		POOL& operator=(const POOL&);
	
		class PRIV;
		PRIV* m_pPriv;
		friend class JOB;
		friend class POOLJOB;
		friend struct POOLITEM;
		friend class QUEUE;
		friend class WORKER::PRIV;
		#endif // GENERATING_DOXYGEN_OUTPUT
	};


//! Class to manage the list of jobs and job subqueues that are given to POOL to execute.
//! This class is a reference counted class.  Any changes to the queue, such as removal of completed jobs, will change all
//! instances of this queue.  There is no MakeExclusive method because JOB's cannot be copied.  This does allow a set and forget 
//! paradigm for QUEUE's.
class CLASSLIBEXPORT QUEUE {
	public:
		
		enum FLAGS {
			FLAG_Default = 					0x0000,				//!< This queue will execute its jobs when a WORKER is available.
			FLAG_Synchronized =				0x0001				//!< This queue will wait until all other jobs before it are finished.
			};
			
		//! Ctor
		explicit QUEUE (
			FLAGS flags = FLAG_Default
			);

		//! Copy ctor	
		QUEUE (
			const QUEUE& rhs
			);
		
		//! Dtor
		~QUEUE (
			);

		//! Assignment operator
		QUEUE& operator=(
			const QUEUE& rhs
			);
	
		//! Add a JOB to 'this'.
		//! @return The MITHREAD::ID of the job added
		ID AddJob (
			JOB& Job,
			bool AddToFront = false			//!< If 'true', places the job ahead of all other jobs or subqueues on this queue.
			);
			
		//! Add a QUEUE subqueue to 'this'.
		//! @return The MITHREAD::ID of the queue added
		ID AddQueue (
			QUEUE& Queue,						//!< The job subqueue to add to this queue.  It may contains JOB's or QUEUE's.
			bool AddToFront = false			//!< If 'true', places the subqueue ahead of all other jobs or subqueues on this queue.
			);
	
		//! Get the flags set for 'this'
		//! @return The set of flags for this queue
		FLAGS GetFlags (
			) const;

		//! Get the ID for 'this'.
		//! @return ID for this queue
		ID GetID (
			) const;

		//! Determine if queue is empty.
		//! Should only be used on queue that has not been added to another queue.
		bool IsEmpty (
			) const;
			
		//! Set a delegate that gets called when all the jobs are complete.
		void SetDelegateOnCompletion (
			DELEGATE_VOID_ID delegate
			);
			
	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		class PRIV;
		PRIV* m_pPriv;
		friend class JOB;
		friend struct QUEUEITEM;
		friend struct QUEUEITEMLIST;
		friend class POOL::PRIV;
		#endif // GENERATING_DOXYGEN_OUTPUT
	};

//! Thread job class to manager specific aspects of a job, such as the error state and completion notification.
class CLASSLIBEXPORT JOB {
	public:
		enum FLAGS {
			FLAG_Default = 					0x0000,	//!< Default, not synchronized, don't delete when complete, don't stop if error.
			FLAG_Synchronized =				0x0001,	//!< This job will wait until all other jobs before it are finished.
			FLAG_DeleteWhenCompleted = 	0x0002,	//!< When completed or cancelled, this job will delete itself, derived class MUST be allocated with new.
			FLAG_StopIfError =				0x0004	//!< Don't run remaining jobs in same queue if get error.
			};
	
		//! Constructor.
		explicit JOB (
			FLAGS flags = FLAG_Default,
			MISTATUSMANAGER* Manager = 0							//!< Status manager to assign to this job, overrides the default one in WORKER and POOL
			);

		//! Destructor.
		virtual ~JOB (
			) = 0;

		//! Get the error state for the thread instance generated by the delegate's return value
		//! There exists an ERRORSTATE for each thread created to prevent a thread from clobbering 
		//! the error state of a different thread.  Therefore to transfer the error state from one
		//! thread to another, we use the ERRORSTATE class.
		//! @return Instance of ERRORSTATE class that 'this' created.
		const ERRORSTATE& GetErrorState (
			) const { return (m_ErrorState); }

		//! Get the flags set for 'this'
		//! @return The set of flags for this job
		FLAGS GetFlags (
			) const { return (m_JobFlags); }

		//! Get the ID for 'this'.
		//! @return ID for this job
		ID GetID (
			) const { return (m_ID); }

		//! Set a delegate that gets called when this job is complete.
		void SetDelegateOnCancelBeforeRun (
			DELEGATE_VOID_ID delegate
			) { m_DelegateOnCancelBeforeRun = delegate; }
			
		//! Set a delegate that gets called when this job is complete.
		void SetDelegateOnCompletion (
			DELEGATE_VOID_ID delegate
			) { m_DelegateOnCompletion = delegate; }
			
	protected:

		//! Get the WORKER instance that this job is running in.
		//! @return Pointer to WORKER if the job is running, NULL if not.
		WORKER* GetWorker (
			) { return (m_Worker); } 

		//! Check if the WORKER::CancelJob() has been called.
		//! @return EUserCancel if the job is canceled, '0' if not.			
		ERRVALUE CheckCancel (
			);
			
	private:

		//! Called if job is cancelled before v_RunJob would be called.
		virtual void v_OnCancelBeforeRun ();
		
		//! Do the actual job.
		//! @return Error code that the job generated or zero if no error.  If an error is generated, the error state can be retrieved via GetErrorState().
		virtual ERRVALUE v_RunJob () = 0;

		#ifndef GENERATING_DOXYGEN_OUTPUT
		JOB (const JOB&);
		JOB& operator=(const JOB&);

		void NotifyCancelBeforeRun ();
		ERRVALUE RunJob ();
		
		ID m_ID;
		FLAGS m_JobFlags;
		ERRORSTATE m_ErrorState;
		DELEGATE_VOID_ID m_DelegateOnCancelBeforeRun;
		DELEGATE_VOID_ID m_DelegateOnCompletion;
		QUEUE::PRIV* m_ParentQueuePriv;
		WORKER* m_Worker;
		MISTATUSMANAGER* m_pManager;
		
		friend struct QUEUEITEM;
		friend class QUEUE::PRIV;
		friend class WORKER::PRIV;
		friend class POOLJOB;
		friend class POOL::PRIV;
		#endif // GENERATING_DOXYGEN_OUTPUT
	};


//! Thread class that MUST be allocated via 'new' and runs detached from the thread that created it.
//! 'Detached' means that the thread cannot be wait'ed upon for completion, or controlled by other threads.
//! This thread will 'delete' itself when the v_ExecuteThread() returns.
class CLASSLIBEXPORT WORKER_DETACHED {
	protected:
		//! Default ctor.  This cannot start the thread because the derived class ctor has not been called yet.
		//! The derived class MUST call 'ExecuteThread()'!
		WORKER_DETACHED ();
		virtual ~WORKER_DETACHED () = 0;
		
		//! This method starts the detached thread.  This MUST be called by the derived class AFTER
		//! all initialization is done in the derived class ctor.
		void ExecuteThread (
			MISTATUSMANAGER* manager = 0,		//!< Set to provide status data or a method to cancel the detached thread's job
			const char* DebugName = 0		//!< Name to assign to the thread, only applicable to Windows debugging under Visual Studio.
			);
		
	private:
		//! Override this method to run the code in the other thread
		virtual void v_ExecuteThread () = 0;
		
		#ifndef GENERATING_DOXYGEN_OUTPUT
		MISTATUSMANAGER* m_pManager;
		const char* m_pDebugName;
		#if defined(WIN32)
			friend UINT32 __stdcall MyThreadFunc (void *vpwd);
		#else
			friend void* MyThreadFunc (void *vpwd);
		#endif
		#endif // GENERATING_DOXYGEN_OUTPUT
	};

}	// End of MITHREAD namespace

#ifndef GENERATING_DOXYGEN_OUTPUT
DEFINE_ENUM_OP_BITWISE(MITHREAD::JOB::FLAGS)
DEFINE_ENUM_OP_BITWISE(MITHREAD::QUEUE::FLAGS)
#endif // GENERATING_DOXYGEN_OUTPUT

#undef	CLASSLIBEXPORT

#endif	//!< INC_MI32_MITHREAD_H
