/**
 * \file rvc/depressionfiller.h
 * \brief Class to fill depressions in elevation raster
 *
 * \if NODOC
 * $Id: depressionfiller.h_v 1.2 2006/05/18 19:24:11 mju Exp $
 *
 * $Log: depressionfiller.h_v $
 * Revision 1.2  2006/05/18 19:24:11  mju
 * Add fillAll variant for processing in-place.
 *
 * Revision 1.1  2006/05/17 17:02:56  mju
 * Initial revision
 *
 * \endif
**/

#ifndef	INC_RVC_DEPRESSIONFILLER_H
#define	INC_RVC_DEPRESSIONFILLER_H

#ifndef	INC_MI32_SIMPLEAR_H
#include <mi32/simplear.h>
#endif

namespace RVC {

class OBJITEM;
class RASTER;

namespace RASTEROP {

//! Class to fill depressions in elevation raster.
//! Provides methods for automatically filling all depressions and for filling individual depression.
class DEPRESSIONFILLER {
	public:

		//! Constructor.
		DEPRESSIONFILLER (
			);

		//! Destructor.
		~DEPRESSIONFILLER (
			);

		//! Fill all depressions, subject to constraint settings.
		//! Target raster will be of same type as source, uncompressed, with no pyramid tiers.
		ERRVALUE FillAll (
			const RVC::OBJITEM& ObjItemSrc,		//!< Source raster location
			RVC::OBJITEM& ObjItemTgt,				//!< Target raster location (may not be existing object)
			SIMPLE_ARRAY<LPOINT2D> *pUnfilledPts = 0	//!< Locations of unfilled depression points returned
			);

		//! Fill all depressions (in place), subject to constraint settings.
		ERRVALUE FillAll (
			const RVC::OBJITEM& ObjItem,			//!< Elevation raster location
			SIMPLE_ARRAY<LPOINT2D> *pUnfilledPts = 0	//!< Locations of unfilled depression points returned
			);

	/*
		//! Fill single depression starting at specified coordinates.
		ERRVALUE FillSingle (
			RVC::RASTER& ObjRaster,					//!< Elevation raster
			INT32 row,									//!< Raster row
			INT32 column								//!< Raster column
			);
	*/

		//! Get number of cells forced to fill before setting base for maximum depth constraint.
		INT32 GetForceFillCells (
			) const { return (m_ForceFillCells); }

		//! Get maximum depth.
		double GetMaxFillDepth (
			) const { return (m_MaxFillDepth); }

		//! Get maximum number of cells to fill for single depression.
		INT32 GetMaxFillCells (
			) const { return (m_MaxFillCells); }

		//! Get maximum number of null cells in single depression to fill.
		INT32 GetMaxNullCells (
			) const { return (m_MaxNullCells); }

		//! Get setting for whether to put null cells at bottom location of unfilled depressions.
		bool GetPutNullAtUnfilled (
			) const { return (m_PutNullAtUnfilled); }

		//! Set number of cells to fill before setting base for maximum depth constraint.
		//! This avoids having small deep pit/well in depression unduly limiting fill.
		//! Default is 100 cells.
		void SetForceFillCells (
			UINT32 ForceFillCells					//!< Size to force filling in cells
			) { m_ForceFillCells = ForceFillCells; }

		//! Set maximum depression size to fill.
		//! Default is 10000 cells.
		void SetMaxFillCells (
			UINT32 MaxFillCells						//!< Maximum fill size in cells
			) { m_MaxFillCells = MaxFillCells; }

		//! Set maximum filling depth.
		//! For integer rasters the maximum allowed value is 250.
		//! Default is 20.
		void SetMaxFillDepth (
			double MaxFillDepth						//!< Maximum depth for individual fill step
			) { m_MaxFillDepth = MaxFillDepth; }

		//! Set maximum size of null area which will be filled.
		//! This allows small null areas which can occur in water areas to be filled.
		//! Default is 100 cells.
		void SetMaxNullCells (
			UINT32 MaxNullCells						//!< Maximum null area size in cells
			) { m_MaxNullCells = MaxNullCells; }

		//! Set whether to put null cell in target raster at locations of unfilled depressions.
		//! Default is to not put null cells.
		void SetPutNullAtUnfilled (
			bool PutNullAtUnfilled
			) { m_PutNullAtUnfilled = PutNullAtUnfilled; }

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		double m_MaxFillDepth;
		UINT32 m_ForceFillCells;
		UINT32 m_MaxFillCells;
		UINT32 m_MaxNullCells;
		bool m_PutNullAtUnfilled;
		#endif
	};


}	// End namespace RASTEROP
}	// End namespace RVC

#endif	// INC_RVC_DEPRESSIONFILLER_H
