/**
 * \file rvc/spatfilt.h 
 * \brief interface for the RVC::SPATIALFILTER class.
 *
 * \if NODOC
 * $Id: spatfilt.h_v 1.13 2012/06/11 18:52:17 bcolombe Exp $
 *
 * $Log: spatfilt.h_v $
 * Revision 1.13  2012/06/11 18:52:17  bcolombe
 * Add laplacian
 *
 * Revision 1.12  2007/11/19 20:09:28  bcolombe
 * Added filtertypes UserDefinedSharp and UserDefinedSmooth
 *
 * Revision 1.11  2007/11/15 21:41:30  bcolombe
 * Added FILTERTYPE_UserDefined
 *
 * Revision 1.10  2007/09/06 17:19:02  scowan
 * Read and write virtual methods changed to take objopenparms.
 *
 * Revision 1.9  2005/12/09 17:29:56  scowan
 * Updated to use credentials.
 *
 * Revision 1.8  2005/07/19 17:49:14  scowan
 * Update to use current simple methods.
 *
 * Revision 1.7  2005/01/21 23:09:26  scowan
 * Fixed deprecated method.
 *
 * Revision 1.6  2003/10/07 17:30:38  scowan
 * Object virtual method changes.
 *
 * Revision 1.5  2003/09/30 16:52:06  dwilliss
 * Doxygen
 *
 * Revision 1.4  2003/09/25 21:20:04  dwilliss
 * Fixed doxygen comment
 *
 * Revision 1.3  2003/09/15 13:49:45  fileserver!dwilliss
 * Doxygen
 *
 * Revision 1.2  2001/06/15 16:09:37  scowan
 * Added to RVC namespace.
 *
 * Revision 1.1  2000/11/30 14:29:55  scowan
 * Initial revision
 *
 * \endif
**/

//!
//!	\class RVC::SPATIALFITLER rvc/spatfilt.h
//!	\brief Manages RVC::OBJTYPE_Filter RVC objects.  
//!
//!	The interface currently wraps the MfFilter functions.
//!
//!	RVC Object Types that this class supports: RVC::OBJTYPE_Filter.
//!
//!	Default RVC Object Type for this class: RVC::OBJTYPE_Filter.
//!
//!	Valid parents for this class:
//!		- RVC::OBJTYPE_File
//!		- RVC::OBJTYPE_Folder
//!
//!	Read() and Write() methods inherited from RVC::SIMPLE class.  View RVC::SIMPLE docs for inherited methods

#ifndef INC_RVC_SPATFILT_H
#define INC_RVC_SPATFILT_H

#ifndef INC_RVC_SIMPLE_H
	#include <rvc/simple.h>
#endif

#ifndef INC_MI32_DOUBLEAR_H
	#include <mi32/doublear.h>
#endif

namespace RVC {

class SPATIALFILTER : public RVC::SIMPLE {
	public:

		enum FILTERTYPE {
			FILTERTYPE_LowPass = 1,
			FILTERTYPE_HighPass,
			FILTERTYPE_HighBoost,
			FILTERTYPE_Median,
			FILTERTYPE_Modal,
			FILTERTYPE_StdDev,
			FILTERTYPE_Gradient,
			FILTERTYPE_Sigma,
			FILTERTYPE_Frost,
			FILTERTYPE_LocalContrast,
			FILTERTYPE_Range,
			FILTERTYPE_Olympic,
			FILTERTYPE_Teager,
			FILTERTYPE_Volterra,
			FILTERTYPE_LUM,
			FILTERTYPE_CS,
			FILTERTYPE_Lee,
			FILTERTYPE_MLM,
			FILTERTYPE_PM,
			FILTERTYPE_AMPM,
			FILTERTYPE_WMMR,
			FILTERTYPE_KuanLLMMSE,
			FILTERTYPE_KuanMAP,
			FILTERTYPE_Froelich,
			FILTERTYPE_UserDefinedSharp,
			FILTERTYPE_UserDefinedSmooth,
			FILTERTYPE_Laplacian,
			FILTERTYPE_LaplacianDiagonals
			};

		SPATIALFILTER (
			);

		SPATIALFILTER (
			const RVC::SPATIALFILTER& rhs
			);

		virtual ~SPATIALFILTER (
			);

		RVC::SPATIALFILTER& operator= (
			const RVC::SPATIALFILTER& rhs
			);

		//! Obtain the boost value
		//! @return: Boost value
		double GetBoost (
			) const { return (m_Boost); }

		//! Obtain the center column of the filter
		//! @return: Filter center column
		INT16 GetCenterColumn (
			) const { return (m_CenterColumn); }

		//! Obtain the center line of the filter
		//! @return: Filter center line
		INT16 GetCenterLine (
			) const { return (m_CenterLine); }

		//! Obtain the filter coefficient data
		//! @return: Filter coefficient data
		const DOUBLE_ARRAY<double>& GetFilterData (
			) const { return (m_FilterData); }

		//! Obtain the type of filter
		//! @return: Filter type
		FILTERTYPE GetFilterType (
			) const { return (m_FilterType); }

		//! Obtain the minimum filter input threshold
		//! @return: Minimum filter input threshold
		double GetMinInputThresh (
			) const { return (m_MinInputThresh); }

		//! Obtain the maximum filter input threshold
		//! @return: Maximum filter input threshold
		double GetMaxInputThresh (
			) const {  return (m_MaxInputThresh); }

		//! Obtain the minimum filter output threshold
		//! @return: Minimum filter output threshold
		double GetMinOutputThresh (
			) const { return (m_MinOutputThresh); }

		//! Obtain the maximum filter output threshold
		//! @return: Maximum filter output threshold
		double GetMaxOutputThresh (
			) const { return (m_MaxOutputThresh); }

		//! Obtain the number of columns in the filter
		//! @return: Number of filter columns
		INT16 GetNumColumns (
			) const { return (m_NumFiltColumns); }

		//! Obtain the number of lines in the filter
		//! @return: Number of filter lines
		INT16 GetNumLines (
			) const { return (m_NumFiltLines); }

		//! Obtain the offset value
		//! @return: Offset value
		double GetOffset (
			) const { return (m_Offset); }

		//! Obtain the scale value
		//! @return: Scale value
		double GetScale (
			) const { return (m_Scale); }

		//! Obtain the version of the filter
		//! @return: Filter version number
		UINT8 GetVersion (
			) const { return (m_Version); }

		//! Set the filter boost value
		void SetBoost (
			double boost
			) { m_Boost = boost; }

		//! Set the filter center column
		void SetCenterColumn (
			INT16 CenterColumn
			) { m_CenterColumn = CenterColumn; }

		//! Set the filter center line
		void SetCenterLine (
			INT16 CenterLine
			) { m_CenterLine = CenterLine; }

		//! Set the filter coefficient data
		void SetFilterData (
			const DOUBLE_ARRAY<double>& FilterData
			) { m_FilterData = FilterData; }

		//! Set the filter type
		void SetFilterType (
			FILTERTYPE  FilterType
			) { m_FilterType = FilterType; }

		//! Set the filter minimum input threshold
		void SetMinInputThresh (
			double MinInputThresh
			) { m_MinInputThresh = MinInputThresh; }

		//! Set the filter maximum input threshold
		void SetMaxInputThresh (
			double MaxInputThresh
			) { m_MaxInputThresh = MaxInputThresh; }

		//! Set the filter minimum output threshold
		void SetMinOutputThresh (
			double MinOutputThresh
			) { m_MinOutputThresh = MinOutputThresh; }

		//! Set the filter maximum output threshold
		void SetMaxOutputThresh (
			double MaxOutputThresh
			) { m_MaxOutputThresh = MaxOutputThresh; }

		//! Set the number of columns in the filter
		void SetNumColumns (
			INT16 NumColumns
			) { m_NumFiltColumns = NumColumns; }

		//! Set the number of lines in the filter
		void SetNumLines (
			INT16 NumLines
			) { m_NumFiltLines = NumLines; }

		//! Set the filter offset value
		void SetOffset (
			double offset
			) { m_Offset = offset; }

		//! Set the filter scale value
		void SetScale (
			double scale
			) { m_Scale = scale; }

		//! Set the filter version
		void SetVersion (
			UINT8 Version
			) { m_Version = Version; }

	protected:

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT

		//! Implementation of RVC::OBJECT virtual methods
		virtual OBJTYPE v_GetDftObjectType () const;

		//! Implementation of RVC::SIMPLE virtual methods
		virtual ERRVALUE v_GetDftNameDesc (const OBJECT& parent, DESCRIPTOR& Descriptor);
		virtual ERRVALUE v_MakeData (RVC::OBJMAKEPARMS& ObjMakeParms);
		virtual ERRVALUE v_ReadData (RVC::OBJOPENPARMS& ObjOpenParms);
		virtual ERRVALUE v_WriteData (RVC::OBJOPENPARMS& ObjOpenParms);

		INT16 m_NumFiltLines;
		INT16 m_NumFiltColumns;
		INT16 m_CenterLine;
		INT16 m_CenterColumn;
		double m_Boost;
		double m_Scale;
		double m_Offset;
		double m_MinInputThresh;
		double m_MaxInputThresh;
		double m_MinOutputThresh;
		double m_MaxOutputThresh;
		UINT8 m_Version;
		FILTERTYPE m_FilterType;
		DOUBLE_ARRAY<double> m_FilterData;
		#endif // GENERATING_DOXYGEN_OUTPUT

	};

}		//! End of RVC namespace

#endif 	//!< INC_RVC_SPATFILT_H
