/**
 * \file spatmod/imagebuffer.h
 * \brief Spatial Modelling (SPATMOD) IMAGE::BUFFER class definition
 *
 * \if NODOC
 * $Id: imagebuffer.h_v 1.16 2011/10/11 14:00:45 mju Exp $
 *
 * $Log: imagebuffer.h_v $
 * Revision 1.16  2011/10/11 14:00:45  mju
 * Remove unneeded hdr.
 *
 * Revision 1.15  2009/07/02 14:53:07  mju
 * In getPtrSampleFast if 64-bit add casts to int64 so properly handles negative values.
 *
 * Revision 1.14  2007/09/20 14:42:13  mju
 * Update docs for getPtrSample.
 *
 * Revision 1.13  2007/02/06 22:26:11  mju
 * Add getTotalRowBytes.
 *
 * Revision 1.12  2007/01/04 17:27:56  mju
 * Add shiftWindowFast.
 *
 * Revision 1.11  2006/11/08 21:26:45  mju
 * Add outputPixelBytes to exportRowBIP.
 *
 * Revision 1.10  2006/11/06 21:39:01  mju
 * Add optional parms to exportRowBIP for skipInvalid and sample mapping.
 *
 * Revision 1.9  2006/10/18 17:42:41  mju
 * Add getPtrSamplesBase.
 *
 * Revision 1.8  2006/07/28 17:28:21  mju
 * Add getPtrSample/IteratorFast.
 *
 * Revision 1.7  2006/07/27 21:48:26  mju
 * Use databuffer for shared buffer with refcount.
 * Support copy ctor and op=.
 * Remove base and const buffers.
 *
 * Revision 1.6  2006/07/14 21:49:37  mju
 * Create buffer_base and buffer_const.
 *
 * Revision 1.5  2006/01/30 17:39:53  mju
 * Move iterator_sample to own hdr and rename.
 *
 * Revision 1.4  2006/01/30 16:41:39  mju
 * Add getINT32_Floor.
 *
 * Revision 1.3  2005/08/19 13:21:59  mju
 * Add getNumRowBytes.
 *
 * Revision 1.2  2005/08/18 14:05:20  mju
 * Add getNumSampleBytes.
 *
 * Revision 1.1  2005/08/16 15:08:52  mju
 * Initial revision
 * \endif
**/

#ifndef	INC_SPATMOD_IMAGEBUFFER_H
#define	INC_SPATMOD_IMAGEBUFFER_H

#ifndef	INC_SPATMOD_IMAGEPIXEL_H
#include <spatmod/imagepixel.h>
#endif

#ifndef	INC_SPATMOD_LOGICALOP_H
#include <spatmod/logicalop.h>
#endif

namespace SPATMOD {

class DATABUFFER;

namespace IMAGE {

class SAMPLEITERATOR;
class SAMPLEITERATOR_CONST;

//=====================================================================================================================

//! Buffer to contain data for an image.
class BUFFER {
	public:

		enum COPYFLAGS {
			COPYFLAG_None =			0x0000,
			COPYFLAG_KeepValid =		0x0001,	//!< Don't replace samples in target which are already valid
			COPYFLAG_NoInvalid =		0x0002,	//!< Don't copy samples from source which are invalid
			};

		//! Default contructor.
		//! Buffer will not be usable until Create() method is called.
		BUFFER (
			);

		//! Copy contructor.
		//! If source buffer has been created then copy will reference same memory.
		BUFFER (
			const BUFFER& rhs
			);

		//! Destructor.
		~BUFFER ();

		//! Assignment.
		//! If source buffer has been created then copy will reference same memory.
		BUFFER& operator= (
			const BUFFER& rhs
			);

		//! Check if rectangle is valid.
		ERRVALUE CheckRectangle (
			INT32 StartColumn,
			INT32 StartRow,
			UINT32 NumColumns,
			UINT32 NumRows
			) const;

		//! Copy rectangle from another buffer.
		ERRVALUE CopyFrom (
			INT32 TgtMinColumn,
			INT32 TgtMinRow,
			const BUFFER& SrcBuffer,
			INT32 SrcMinColumn,
			INT32 SrcMinRow,
			UINT32 SrcNumColumns,
			UINT32 SrcNumRows
			);

		//! Create buffer with specified pixel properties, size and optional external data buffers.
		//! If data buffers are provided, the BUFFER will not take ownership.  Specified data buffers
		//! must also point to contiguous area of memory of (at least) NumColumns * NumRows * [bytes per sample]
		//! bytes with suitable alignment based on sample type.  Data buffers must remain valid until this
		//! referencing BUFFER is destroyed.
		//! If no 'validity' is specified the initial validity will be true if 'data' specified and false if not.
		ERRVALUE Create (
			const PIXEL& PixelProperties,
			UINT32 NumColumns,
			UINT32 NumRows,
			void **data = 0,
			UINT8 *validity = 0
			);

		//! Create buffer as 'window' into existing buffer.
		//! Window must be within limits of existing buffer or error will be returned
		//! Existing buffer may be destroyed after returns as actual data is reference counted.
		ERRVALUE Create (
			BUFFER& existing,						//!< Existing buffer to reference
			INT32 ColOffset,						//!< Column offset relative to existing buffer
			INT32 RowOffset,						//!< Row offset relative to existing buffer
			UINT32 WinColumns,					//!< Number of columns in window
			UINT32 WinRows							//!< Number of rows in window
			);

		//! Create buffer with subset of samples in existing buffer.
		//! Existing buffer may be destroyed after returns as actual data is reference counted.
		//! PixelProperties for new buffer will be derived from specified sample subset.
		ERRVALUE CreateSubset (
			BUFFER& existing,						//!< Existing buffer to reference
			const UINT16 *SampleIdxs,			//!< Sample indices in existing buffer
			UINT16 NumSamples,					//!< Number of sample indices
			bool OwnValidity						//!< True to keep own validity data, false to use from existing buffer
			);

		//! Destroy buffer.
		void Destroy (
			);

		//! Export row of pixels to external BIP buffer.
		//! No special processing for invalid pixels.
		//! Allows a translation array to select which sample is mapped to each output sample position.
		ERRVALUE ExportRowBIP (
			void *target,							//!< Target buffer, will be packed as band interleaved by pixel, no padding for alignment
			UINT32 row,								//!< Row number
			UINT32 StartColumn = 0,				//!< Starting column
			UINT32 NumColumns = 0,				//!< Number of column, 0 for number of remaining columns in row
			bool SkipInvalid = false,			//!< Skip invalid pixels, leaving unchanged in target buffer
			const UINT16 *SampleTrans = 0,	//!< Sample translation array, NULL if not used
			UINT16 NumSamplesTrans = 0,		//!< Number of entries in sample translation array
			int OutputPixelBytes = 0			//!< Number of bytes per output pixel, 0 to derive from buffer
			) const;

		//! Export row of pixels to external BIP buffer with specified null pixel.
		ERRVALUE ExportRowBIP (
			void *target,							//!< Target buffer, will be packed as band interleaved by pixel, no padding for alignment
			const PIXEL& nullpixel,				//!< Pixel value to export non-valid pixels as
			UINT32 row,								//!< Row number
			UINT32 StartColumn = 0,				//!< Starting column
			UINT32 NumColumns = 0				//!< Number of column, 0 for number of remaining columns in row
			) const;

		//! Get SAMPLEITERATOR_CONST starting at specified sample location.
		ERRVALUE GetIterator (
			SAMPLEITERATOR_CONST& iterator,	//!< Iterator returned
			UINT16 SampleNum,						//!< Sample index
			INT32 row,								//!< Row number
			INT32 column = 0						//!< Column number
			) const;

		//! Get SAMPLEITERATOR starting at specified sample location.
		ERRVALUE GetIterator (
			SAMPLEITERATOR& iterator,			//!< Iterator returned
			UINT16 SampleNum,						//!< Sample index
			INT32 row,								//!< Row number
			INT32 column = 0						//!< Column number
			);

		//! Get number of columns in buffer.
		UINT32 GetNumColumns (
			) const { return (m_WinNumColumns); }

		//! Get number of rows in buffer.
		UINT32 GetNumRows (
			) const { return (m_WinNumRows); }

		//! Get number of bytes in (window) row.
		UINT32 GetNumRowBytes (
			UINT16 SampleNum
			) const { return (m_WinNumColumns * m_SampleBytes[SampleNum]); }

		//! Get number of bytes per specified sample.
		UINT32 GetNumSampleBytes (
			UINT16 SampleNum
			) const { return (m_SampleBytes[SampleNum]); }

		//! Get number of samples.
		//! Same as GetPixelProperties().GetNumSamples().
		UINT16 GetNumSamples (
			) const { return (m_PixelProperties.GetNumSamples()); }

		//! Get pixel properties for buffer.
		const PIXEL& GetPixelProperties (
			) const { return (m_PixelProperties); }

		//! Get pointer to sample.
		//! If sample index is >= number of samples, a pointer to the validity buffer position will be returned.
		//! This is useful when have additional validity samples.
		//! Note that samples for row are always contiguous in memory, and for efficiency this
		//! method should therefore not be used to iterate through the samples in a single row.
		//! @return NULL if any parameter out of range.
		const void *GetPtrSample (
			UINT16 SampleNum,						//!< Sample index, UINT16_MAX for validity
			INT32 row,								//!< Row number
			INT32 column = 0						//!< Column number
			) const;

		//! Get pointer to sample.
		//! If sample index is >= number of samples, a pointer to the validity buffer position will be returned.
		//! This is useful when have additional validity samples.
		//! Note that samples for row are always contiguous in memory, and for efficiency this
		//! method should therefore not be used to iterate through the samples in a single row.
		//! @return NULL if any parameter out of range.
		void *GetPtrSample (
			UINT16 SampleNum,						//!< Sample index, UINT16_MAX for validity
			INT32 row,								//!< Row number
			INT32 column = 0						//!< Column number
			);

		//! Get pointer to sample with no validation.
		//! Note that samples for row are always contiguous in memory, and for efficiency this
		//! method should therefore not be used to iterate through the samples in a single row.
		const void *GetPtrSampleFast (
			UINT16 SampleNum,						//!< Sample index
			INT32 row,								//!< Row number
			INT32 column = 0						//!< Column number
			) const {
			#ifdef PTRS_ARE_64_BIT
				return (static_cast<const UINT8*>(m_TotSampleBase[SampleNum]) + (row + static_cast<INT64>(m_WinOffRow)) * static_cast<INT64>(m_RowBytes[SampleNum]) + (column + static_cast<INT64>(m_WinOffColumn)) * static_cast<INT64>(m_SampleBytes[SampleNum]));
			#else
				return (static_cast<const UINT8*>(m_TotSampleBase[SampleNum]) + (row + m_WinOffRow) * m_RowBytes[SampleNum] + (column + m_WinOffColumn) * m_SampleBytes[SampleNum]);
			#endif
			}

		//! Get pointer to sample with no validation.
		//! Note that samples for row are always contiguous in memory, and for efficiency this
		//! method should therefore not be used to iterate through the samples in a single row.
		void *GetPtrSampleFast (
			UINT16 SampleNum,						//!< Sample index
			INT32 row,								//!< Row number
			INT32 column = 0						//!< Column number
			) {
			#ifdef PTRS_ARE_64_BIT
				return (static_cast<UINT8*>(m_TotSampleBase[SampleNum]) + (row + static_cast<INT64>(m_WinOffRow)) * static_cast<INT64>(m_RowBytes[SampleNum]) + (column + static_cast<INT64>(m_WinOffColumn)) * static_cast<INT64>(m_SampleBytes[SampleNum]));
			#else
				return (static_cast<UINT8*>(m_TotSampleBase[SampleNum]) + (row + m_WinOffRow) * m_RowBytes[SampleNum] + (column + m_WinOffColumn) * m_SampleBytes[SampleNum]);
			#endif
			}

		//! Get array of pointers to base sample buffers.
		void ** GetPtrSamplesBase (
			) { return (m_TotSampleBase); }

		//! Get pointer to validity mask.
		//! Note that validity values for row are always contiguous in memory, and for efficiency this
		//! method should therefore not be used to iterate through the validity for a single row.
		//! @return NULL if any parameter out of range.
		const UINT8 *GetPtrValidity (
			INT32 row,								//!< Row number
			INT32 column = 0						//!< Column number
			) const;

		//! Get pointer to validity mask.
		//! Note that validity values for row are always contiguous in memory, and for efficiency this
		//! method should therefore not be used to iterate through the validity for a single row.
		//! @return NULL if any parameter out of range.
		UINT8 *GetPtrValidity (
			INT32 row,								//!< Row number
			INT32 column = 0						//!< Column number
			);

		//! Get pointer to validity mask with no validation.
		//! Note that validity values for row are always contiguous in memory, and for efficiency this
		//! method should therefore not be used to iterate through the validity for a single row.
		const UINT8 *GetPtrValidityFast (
			INT32 row,								//!< Row number
			INT32 column = 0						//!< Column number
			) const { return (m_TotValidityBase + (row + m_WinOffRow) * m_TotNumColumns + column + m_WinOffColumn); }

		//! Get pointer to validity mask with no validation.
		//! Note that validity values for row are always contiguous in memory, and for efficiency this
		//! method should therefore not be used to iterate through the validity for a single row.
		UINT8 *GetPtrValidityFast (
			INT32 row,								//!< Row number
			INT32 column = 0						//!< Column number
			) { return (m_TotValidityBase + (row + m_WinOffRow) * m_TotNumColumns + column + m_WinOffColumn); }

		//! Get total number of columns.
		//! Used for specialized windowed processing.
		UINT32 GetTotalColumns (
			) const { return (m_TotNumColumns); }

		//! Get total number of rows.
		//! Used for specialized windowed processing.
		UINT32 GetTotalRows (
			) const { return (m_TotNumRows); }

		//! Get total number of bytes in row.
		//! Used for specialized processing.
		UINT32 GetTotalRowBytes (
			UINT16 SampleNum
			) const { return (m_RowBytes[SampleNum]); }

		//! Get window offset column.
		//! Used for specialized windowed processing.
		UINT32 GetWindowColumn (
			) const { return (m_WinOffColumn); }

		//! Get window offset row.
		//! Used for specialized windowed processing.
		UINT32 GetWindowRow (
			) const { return (m_WinOffRow); }

		//! Import row of pixels from external BIP buffer.
		ERRVALUE ImportRowBIP (
			const void *source,					//!< Source buffer, packed as band interleaved by pixel, no padding for alignment
			const UINT8 *validity,				//!< Validity buffer, NULL if all pixels valid
			UINT32 row,								//!< Row number
			UINT32 StartColumn = 0,				//!< Starting column
			UINT32 NumColumns = 0				//!< Number of column, 0 for number of remaining columns in row
			);

		//! Import row of pixels from external BIP buffer.
		ERRVALUE ImportRowBIP (
			const void *source,					//!< Source buffer, packed as band interleave by pixel, no padding for alignment
			const PIXEL& nullpixel,				//!< Null pixel value
			UINT32 row,								//!< Row number
			UINT32 StartColumn = 0,				//!< Starting column
			UINT32 NumColumns = 0				//!< Number of column, 0 for number of remaining columns in row
			);

		//! Determine if entire buffer for each sample is contiguous in memory.
		//! Note that individual rows are alway contiguous.
		bool IsContiguous (
			) const { return (m_IsContiguous); }

		//! Determine if buffer has been created yet.
		bool IsCreated (
			) const;

		//! Set rectangular portion of buffer to be invalid.
		ERRVALUE SetInvalid (
			INT32 StartColumn,					//!< Starting column in buffer
			INT32 StartRow,						//!< Starting row
			UINT32 NumColumns,					//!< Number of columns
			UINT32 NumRows							//!< Number of rows
			);

		//! Set validity for all pixels in buffer window.
		void SetValidityAll (
			bool validity							//!< Validity to set, true for valid, false for not
			);

		//! Set validity for all pixels using logical operation with validity from specified buffer.
		//! The LOGICALOP_And, LOGICALOP_Or are supported.
		//! Both buffers must have the same window dimensions, but may have different pixel properties.
		ERRVALUE SetValidityLogical (
			LOGICALOP operation,					//!< Logical operation to perform
			const BUFFER& buffer					//!< Buffer to use in operation
			);

		//! Shift window position by specified deltas without validation.
		void ShiftWindowFast (
			INT32 DeltaColumn,
			INT32 DeltaRow
			) { m_WinOffColumn += DeltaColumn; m_WinOffRow += DeltaRow; }

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT

		DATABUFFER *m_pSampleBuffer;
		DATABUFFER *m_pValidityBuffer;
		void **m_TotSampleBase;
		UINT8 *m_TotValidityBase;
		UINT32 *m_SampleBytes;
		UINT32 *m_RowBytes;
		PIXEL m_PixelProperties;
		UINT32 m_TotNumRows;
		UINT32 m_TotNumColumns;
		UINT32 m_WinNumRows;
		UINT32 m_WinNumColumns;
		UINT32 m_WinOffRow;
		UINT32 m_WinOffColumn;
		bool m_IsContiguous;

		void CopyRef (const BUFFER& rhs);

		#endif // GENERATING_DOXYGEN_OUTPUT
	};

#ifndef GENERATING_DOXYGEN_OUTPUT
DEFINE_ENUM_OP_BITWISE(BUFFER::COPYFLAGS);
#endif

//=====================================================================================================================

}	// End namespace IMAGE
}	// End namespace SPATMOD

#endif	 // INC_SPATMOD_IMAGEBUFFER_H
