/**
 * \file spatmod/imagefilterclassify.h
 * \brief Spatial Modelling (SPATMOD) IMAGE::FILTER_CLASSIFY definition.
 *
 * \if NODOC
 * $Id: imagefilterclassify.h_v 1.5 2013/01/07 18:23:56 mju Exp $
 *
 * $Log: imagefilterclassify.h_v $
 * Revision 1.5  2013/01/07 18:23:56  mju
 * Keep track of which classes are used via bitset and use to manage stats.
 *
 * Revision 1.4  2013/01/04 21:32:56  mju
 * Add methods for training.
 *
 * Revision 1.3  2013/01/02 17:33:47  mju
 * Add getCountUnclassified.
 *
 * Revision 1.2  2013/01/02 15:42:12  mju
 * Specify min/max classnum in ctor, add setCalcStats, setValueUnclassified.
 *
 * Revision 1.1  2012/12/31 18:52:47  mju
 * Initial revision
 * \endif
**/

#ifndef	INC_SPATMOD_IMAGEFILTERCLASSIFY_H
#define	INC_SPATMOD_IMAGEFILTERCLASSIFY_H

#ifndef	INC_SPATMOD_IMAGEFILTER_H
#include <spatmod/imagefilter.h>
#endif

#ifndef	INC_MI32_BITSET_H
#include <mi32/bitset.h>
#endif


#ifndef GENERATING_DOXYGEN_OUTPUT
// Forward declarations.
class STAT_MULTIVAR;
namespace SPATMOD {
	namespace IMAGE {
		class SAMPLEITERATOR_CONST;
		}
	}
#endif

namespace SPATMOD {
namespace IMAGE {

//=====================================================================================================================

//! Base filter for classification.
//! Provides ability to optionally compute statistics while classifying.
class FILTER_CLASSIFY : public FILTER {
	public:

		//! Destructor.
		virtual ~FILTER_CLASSIFY ();

		//! Get statistics for specified class number.
		//! @return Statistics, results indeterminate if statistics not computed or class number out of range
		//! Note that statistics are reset each time the image is processed (via PullBegin).
		const STAT_MULTIVAR& GetClassStats (
			UINT32 ClassNum							//!< Class number
			) const;

		//! Get count of "unclassified" cells.
		//! @return Count, will be 0 if statistics are not computed.
		UINT64 GetCountUnclassified () const
			{ return (m_CountUnclassified); }

		//! Get maximum defined class number.
		//! Result will only be valid after classifier is "trained".
		UINT32 GetMaxClassNum () const
			{ return (m_MinClassNum + m_ClassUsedSet.GetNumEntries() - 1); }

		//! Get minimum defined class number.
		UINT32 GetMinClassNum () const
			{ return (m_MinClassNum); }

		//! Get value to use for "unclassified" result.
		UINT32 GetValueUnclassified () const
			{ return (m_ValueUnclassified); }

		//! Set whether to calculate statistics during processing.
		//! Default is to not calculate, if used must be before Initialize called.
		void SetCalcStats (
			bool CalcStats = true
			) { m_CalcStats = CalcStats; }

		//! Set value to use for "unclassified" result.
		//! The default value will be determined based on the minimum defined class number.  If this value is
		//! > 0 the value for unclassified will be 0, otherwise it will be the maximum allowed for the DATATYPE
		//! as determined by the maximum defined class number.
		void SetValueUnclassified (
			UINT32 ValueUnclassified
			) { m_ValueUnclassified = ValueUnclassified; }

		//! Perform automatic training.
		ERRVALUE TrainAutomatic (
			UINT16 SampleInterval					//!< Cell sampling interval
			);

		//! Perform training from class samples.
		ERRVALUE TrainFromSamples (
			STAGE *pStageSamples						//!< Stage for class samples, must match input stage dimensions
			);

	protected:

		//! Construct with specified minimum class number.
		FILTER_CLASSIFY (
			STAGE *pStageInput,						//!< Pointer to stage
			bool TakeOwnership,						//!< Filter takes ownership of previous stage and will delete when needed
			UINT32 MinClassNum						//!< Minimum class number
			);

		//! Construct with specified minimum class number and set of classes used.
		FILTER_CLASSIFY (
			STAGE *pStageInput,						//!< Pointer to stage
			bool TakeOwnership,						//!< Filter takes ownership of previous stage and will delete when needed
			UINT32 MinClassNum,						//!< Minimum class number
			const BITSET& ClassUsedSet				//!< Set of classes used
			);

		//! Access set of classes used.
		//! @return BITSET, first position corresponds to minimum class number.
		BITSET& AccessClassUsedSet ()
			{ return (m_ClassUsedSet); }

		//! Get set of classes used.
		//! @return BITSET, first position corresponds to minimum class number.
		const BITSET& GetClassUsedSet () const
			{ return (m_ClassUsedSet); }

		//! Set set of classes used.
		//! @return BITSET, first position corresponds to minimum class number.
		void SetClassUsedSet (
			const BITSET& ClassUsedSet
			) { m_ClassUsedSet = ClassUsedSet; }

		//! Set minimum class number.
		void SetMinClassNum (
			UINT32 MinClassNum
			) { m_MinClassNum = MinClassNum; }

		//! Perform initialization.
		//! If override, must call base method before performing own initialization.
		virtual ERRVALUE v_Initialize ();

		//! Begin pull of specific section.
		//! If override, must must call base method.
		virtual ERRVALUE v_PullBegin (const SECTION_ITERATOR& iterator);

		//! Finish pulling subsections.
		//! If override, must must call base method.
		virtual ERRVALUE v_PullEnd ();

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		typedef FILTER BASECLASS;

		SIMPLE_ARRAY<STAT_MULTIVAR*> m_ClassStats;
		double *m_SrcValues;
		void const ** m_pSrcValue;
		SAMPLEITERATOR_CONST *m_itSrcValue;
		UINT32 *m_TgtValue;
		BITSET m_ClassUsedSet;
		UINT64 m_CountUnclassified;
		UINT32 m_MinClassNum;
		UINT32 m_ValueUnclassified;
		bool m_CalcStats;

		void AttachItPtrSrcValue (BUFFER& BufSrc, UINT32 row);
		void ProcessTgtRow (void *pTgtValue, UINT32 NumColumns) const;
		void SetupItPtrSrcValue ();

		ERRVALUE v_PullSubSection (const SECTION& SubSection, BUFFER& BufTgt);

		// Unimplemented
		FILTER_CLASSIFY (const FILTER_CLASSIFY&);
		FILTER_CLASSIFY& operator= (const FILTER_CLASSIFY&);
		#endif // GENERATING_DOXYGEN_OUTPUT

		//! Perform classification on source values.
		virtual ERRVALUE v_Classify (
			SAMPLEITERATOR_CONST *itSrcValue,	//!< Array of source value iterators
			void const **pSrcValue,					//!< Array of pointers to source values if all have same DATATYPE, for optimization with common DATATYPEs, will be 0 if not same DATATYPEs
			UINT8 *pTgtValidity,						//!< Target validity, initially set based on source validity
			UINT32 *pTgtValue,						//!< Array to hold target values
			UINT32 NumPixels							//!< Number of pixels in span
			) const = 0;

		//! Perform "automatic" training using input values.
		//! Must be implemented if classifier supports automatic training, default implementation returns error.
		virtual ERRVALUE v_TrainAutomatic (
			SAMPLEITERATOR_CONST *itSrcValue,	//!< Array of source value iterators
			void const **pSrcValue,					//!< Array of pointers to source values if all have same DATATYPE, for optimization with common DATATYPEs, will be 0 if not same DATATYPEs
			const UINT8 *pSrcValidity,				//!< Source validity
			UINT32 NumPixels							//!< Number of pixels in span
			);

		//! Perform training using class samples and input values.
		//! Must be implemented if classifier supports training from samples, default implementation returns error.
		virtual ERRVALUE v_TrainFromSamples (
			SAMPLEITERATOR_CONST *itSrcValue,	//!< Array of source value iterators
			void const **pSrcValue,					//!< Array of pointers to source values if all have same DATATYPE, for optimization with common DATATYPEs, will be 0 if not same DATATYPEs
			const UINT8 *pSrcValidity,				//!< Source validity
			const UINT32 *pSampleValue,			//!< Sample values
			const UINT8 *pSampleValidity,			//!< Sample validity
			UINT32 NumPixels							//!< Number of pixels in span
			);

		//! Called at beginning of each training pass.
		virtual ERRVALUE v_TrainPassBegin (
			bool bAutomatic							//!< Whether doing "automatic" training (true) or "from samples" (false)
			);

		//! Called at end of training pass.
		//! Must be implemented if classifier supports training, default implementation returns error.
		//! Implementation must return 0 when training is complete, > 0 if another pass is needed or < 0 if error.
		virtual int v_TrainPassEnd ();

	};

//=====================================================================================================================

}	// End namespace IMAGE
}	// End namespace SPATMOD

#endif	// INC_SPATMOD_IMAGEFILTERCLASSIFY_H
