/**
 * \file imagefilterdatatype.h <spatmod/imagefilterdatatype.h>
 * \brief Spatial Modelling (SPATMOD) IMAGE::FILTER_DATATYPE definition.
 *
 * \if NODOC
 * $Id: imagefilterdatatype.h_v 1.6 2009/12/01 15:37:40 mju Exp $
 *
 * $Log: imagefilterdatatype.h_v $
 * Revision 1.6  2009/12/01 15:37:40  mju
 * Add separate modes for round vs floor.
 *
 * Revision 1.5  2008/09/03 17:00:03  bcolombe
 * Added precision
 *
 * Revision 1.4  2006/01/30 17:40:05  mju
 * Rename sample iterator class.
 *
 * Revision 1.3  2006/01/27 22:11:37  mju
 * Add mode_NoCheck.
 *
 * Revision 1.2  2005/08/18 17:17:13  mju
 * No need for pullbegin/end overrides.
 *
 * Revision 1.1  2005/08/16 15:09:27  mju
 * Initial revision
 *
 * \endif
**/

#ifndef	INC_SPATMOD_IMAGEFILTERDATATYPE_H
#define	INC_SPATMOD_IMAGEFILTERDATATYPE_H

#ifndef	INC_SPATMOD_IMAGEFILTER_H
#include <spatmod/imagefilter.h>
#endif

#ifndef	INC_SPATMOD_IMAGEBUFFER_H
#include <spatmod/imagebuffer.h>
#endif

namespace SPATMOD {
namespace IMAGE {

//=====================================================================================================================

//! Filter to change data type of an image.
class FILTER_DATATYPE : public FILTER {
	public:

		//! Mode for how values are converted and handling for out of range values
		enum MODE {
			MODE_Source =				0,		//!< Use data type and sample values directly from source
			MODE_RoundLimit =			1,		//!< Limit values to target range, rounding if integer result
			MODE_RoundInvalidate =	2,		//!< Invalidate pixels with sample values outside target range, rounding if integer result
			MODE_RoundRescale =		3,		//!< Rescale values to target range, rounding if integer result
			MODE_RoundNoCheck =		4,		//!< Don't check value, only allowed if source type fits target, rounding if integer result
			MODE_FloorLimit =			5,		//!< Limit values to target range, flooring if integer result
			MODE_FloorInvalidate =	6,		//!< Invalidate pixels with sample values outside target range, flooring if integer result
			MODE_FloorRescale =		7,		//!< Rescale values to target range, flooring if integer result
			MODE_FloorNoCheck =		8,		//!< Don't check value, only allowed if source type fits target, flooring if integer result
			MODE_Limit = MODE_RoundLimit,
			MODE_Invalidate = MODE_RoundInvalidate,
			MODE_Rescale = MODE_RoundRescale,
			MODE_NoCheck = MODE_RoundNoCheck,
			};

		//! Filter settings.
		class SETTINGS {
			public:
				//! Default constructor.
				SETTINGS ();
				//! Construct with data type and mode.
				SETTINGS (
					DATATYPE DataType,
					MODE Mode
					);
				//! Construct with data type, mode and range.
				SETTINGS (
					DATATYPE DataType,
					MODE Mode,
					double MinValue,
					double MaxValue
					);
				//! Get mode.
				MODE GetMode (
					) const { return (m_Mode); }
				//! Get data type.
				DATATYPE GetDataType (
					) const { return (m_DataType); }
				//! Get whether to do "floor" when convert to integer.
				bool GetDoFloor (
					) const { return (m_DoFloor); }
				//! Get maximum value.
				double GetMaxValue (
					) const { return (m_MaxValue); }
				//! Get minimum value.
				double GetMinValue (
					) const { return (m_MinValue); }
				//! Get offset.
				double GetOffset (
					) const { return (m_Offset); }
				//! Get precision
				int GetPrecision (
					) const {return (m_Precision); }
				//! Get scale.
				double GetScale (
					) const { return (m_Scale); }
				//! Determine if range is full for data type.
				bool IsRangeFull (
					) const { return (m_IsRangeFull); }
				//! Determine if range is valid for data type.
				bool IsRangeValid (
					) const;
				//! Set data type.
				void SetDataType (
					DATATYPE DataType
					);
				//! Set whether to do "floor" when convert to integer (default is to round).
				void SetDoFloor (
					bool DoFloor = true
					) { m_DoFloor = DoFloor; }
				//! Set mode.
				void SetMode (
					MODE Mode
					) { m_Mode = Mode; }
				//! Set precision.
				void SetPrecision (
					int Precision
					) { m_Precision = Precision; }
				//! Set range.
				void SetRange (
					double MinValue,
					double MaxValue
					);

				#ifndef GENERATING_DOXYGEN_OUTPUT
				// Compute default range based on data type.
				void ComputeRangeDefault (
					);
				// Compute scale/offset for MODE_Round/FloorRescale.
				void ComputeScaleOffset (
					double SrcMinValue,
					double SrcMaxValue
					);
				// Process samples 
				void Process (
					SAMPLEITERATOR& itTgt,
					UINT8 *TgtValidity,
					SAMPLEITERATOR& itSrc,
					INT32 NumColumns
					) const;
			private:
				double m_MinValue;
				double m_MaxValue;
				double m_Scale;
				double m_Offset;
				int m_Precision;
				MODE m_Mode;
				DATATYPE m_DataType;
				bool m_IsRangeFull;
				bool m_DoFloor;
				#endif // GENERATING_DOXYGEN_OUTPUT
			};

		//! Construct with same target data type and mode for all samples and default range.
		FILTER_DATATYPE (
			STAGE *pSourceImage,						//!< Source image stage
			bool TakeOwnership,						//!< Filter takes ownership of source stage and will delete when needed
			DATATYPE TargetDataType,				//!< Target data type
			MODE Mode,									//!< Mode for value conversion and range validation
			bool PassOpacity = true					//!< Pass opacity samples unchanged
			);

		//! Construct with same target data type and mode for all samples and specified range.
		FILTER_DATATYPE (
			STAGE *pSourceImage,						//!< Source image stage
			bool TakeOwnership,						//!< Filter takes ownership of source stage and will delete when needed
			DATATYPE TargetDataType,				//!< Target data type
			MODE Mode,									//!< Mode for value conversion and range validation
			double MinValue,							//!< Minimum target sample value
			double MaxValue,							//!< Maximum target sample value
			bool PassOpacity = true					//!< Pass opacity samples unchanged
			);

		//! Construct with same target data type and mode for all samples and default range.
		FILTER_DATATYPE (
			STAGE *pSourceImage,						//!< Source image stage
			bool TakeOwnership,						//!< Filter takes ownership of source stage and will delete when needed
			DATATYPE TargetDataType,				//!< Target data type
			int Precision,							//!< Sample Precision
			MODE Mode,									//!< Mode for value conversion and range validation
			bool PassOpacity = true					//!< Pass opacity samples unchanged
			);

		//! Construct with same target data type and mode for all samples and specified range.
		FILTER_DATATYPE (
			STAGE *pSourceImage,						//!< Source image stage
			bool TakeOwnership,						//!< Filter takes ownership of source stage and will delete when needed
			DATATYPE TargetDataType,				//!< Target data type
			int Precision,							//!< Sample Precision
			MODE Mode,									//!< Mode for value conversion and range validation
			double MinValue,							//!< Minimum target sample value
			double MaxValue,							//!< Maximum target sample value
			bool PassOpacity = true					//!< Pass opacity samples unchanged
			);

		//! Construct with SETTINGS array.
		FILTER_DATATYPE (
			STAGE *pSourceImage,						//!< Source image stage
			bool TakeOwnership,						//!< Filter takes ownership of source stage and will delete when needed
			const SETTINGS *Settings				//!< Array of settings, one entry for each input sample
			);

		// Destructor.
		virtual ~FILTER_DATATYPE ();

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT
		SETTINGS *m_Settings;						// Settings indexed by sample number

		// STAGE/FILTER overrides.
		virtual ERRVALUE v_Initialize ();
		virtual ERRVALUE v_PullSubSection (const SECTION& SubSection, BUFFER& buffer);

		// Unimplemented
		FILTER_DATATYPE (const FILTER_DATATYPE&);
		FILTER_DATATYPE& operator= (const FILTER_DATATYPE&);
		#endif // GENERATING_DOXYGEN_OUTPUT

	};

//=====================================================================================================================

}	// End namespace IMAGE
}	// End namespace SPATMOD

#endif	// INC_SPATMOD_IMAGEFILTERDATATYPE_H

