/**
 * \file spatmod/imagefilteroutline.h
 * \brief Spatial Modelling (SPATMOD) IMAGE::FILTER_OUTLINE definition.
 *
 * \if NODOC
 * $Id: imagefilteroutline.h_v 1.1 2012/12/11 18:58:00 mju Exp $
 *
 * $Log: imagefilteroutline.h_v $
 * Revision 1.1  2012/12/11 18:58:00  mju
 * Initial revision
 *
 * \endif
**/

#ifndef	INC_SPATMOD_IMAGEFILTEROUTLINE_H
#define	INC_SPATMOD_IMAGEFILTEROUTLINE_H

#ifndef	INC_SPATMOD_IMAGEFILTER_H
#include <spatmod/imagefilter.h>
#endif

#ifndef	INC_SPATMOD_IMAGESAMPLEGROUPLIST_H
#include <spatmod/imagesamplegrouplist.h>
#endif

namespace SPATMOD {
namespace IMAGE {

//=====================================================================================================================

//! Filter to generate an "outline" image.
//! Each pixel will be compared to its 4 orthogonal neighbors and if identical to all it will be set as invalid or
//! its opacity component will be adjusted.  The opacity component itself is not used in the comparison.  However, if
//! multiple opacity components exists, each group of samples in the pixel will be processed separately.  This allows
//! the filter to work on multi-frame stereo images.
class FILTER_OUTLINE : public FILTER {
	public:

		//! Constructor.
		FILTER_OUTLINE (
			STAGE *pPreviousStage,					//!< Pointer to previous stage
			bool TakeOwnership						//!< Filter takes ownership of previous stage and will delete when needed
			);

		//! Destructor.
		virtual ~FILTER_OUTLINE ();

		//! Set to invalidate non-outline cells (instead of adjusting opacity).
		//! By default, cells not included in the outline will be set to 100% transparent (0% opaque) if an alpha component
		//! exists.  If no alpha component exists non-outline cells will always be set as invalid.
		void SetInvalidate (
			bool invalidate = true					//!< Whether to invalidate non-outline cells
			) { m_bInvalidate = invalidate; }

		//! Set opacity factor.
		//! By default, cells not included in the outline will be set to 100% transparent (0% opaque) if an alpha component
		//! exists.  The specified factor will be multiplied by the opacity for the cell, thus a factor of 0 will make the
		//! non-border cell 0% opaque (invisible).  This setting has no effect if no alpha component exists.
		void SetOpacityFactor (
			double factor								//!< Opacity factor (0-1) for making non-border cells partially transparent
			);

	private:
		#ifndef GENERATING_DOXYGEN_OUTPUT

		SAMPLEGROUPLIST m_SampleGroupList;
		bool m_bInvalidate;
		UINT16 m_OpacityFactor;
		UINT8 *m_IsEdge;

		void ProcessUINT8 (UINT16 s, INT32 row, INT32 NumColumns, const BUFFER& SrcBuffer);
		void ProcessSINT8 (UINT16 s, INT32 row, INT32 NumColumns, const BUFFER& SrcBuffer);
		void ProcessUINT16 (UINT16 s, INT32 row, INT32 NumColumns, const BUFFER& SrcBuffer);
		void ProcessSINT16 (UINT16 s, INT32 row, INT32 NumColumns, const BUFFER& SrcBuffer);
		void ProcessUINT32 (UINT16 s, INT32 row, INT32 NumColumns, const BUFFER& SrcBuffer);
		void ProcessSINT32 (UINT16 s, INT32 row, INT32 NumColumns, const BUFFER& SrcBuffer);
		void ProcessFLOAT32 (UINT16 s, INT32 row, INT32 NumColumns, const BUFFER& SrcBuffer);
		void ProcessFLOAT64 (UINT16 s, INT32 row, INT32 NumColumns, const BUFFER& SrcBuffer);

		// BASE/STAGE/FILTER overrides.
		virtual ERRVALUE v_Initialize ();
		virtual ERRVALUE v_PullBegin (const SECTION_ITERATOR& iterator);
		virtual ERRVALUE v_PullEnd ();
		virtual ERRVALUE v_PullSubSection (const SECTION& SubSection, BUFFER& buffer);

		// Unimplemented
		FILTER_OUTLINE (const FILTER_OUTLINE&);
		FILTER_OUTLINE& operator= (const FILTER_OUTLINE&);
		#endif // GENERATING_DOXYGEN_OUTPUT

	};

//=====================================================================================================================

}	// End namespace IMAGE
}	// End namespace SPATMOD

#endif	// INC_SPATMOD_IMAGEFILTEROUTLINE_H

