/**
 * \file spatmod/imagehelperresample.h
 * \brief SPATMOD::IMAGE::HELPER_RESAMPLE definition
 *
 * \if NODOC
 * $Id: imagehelperresample.h_v 1.4 2013/04/04 15:01:35 mju Exp $
 *
 * $Log: imagehelperresample.h_v $
 * Revision 1.4  2013/04/04 15:01:35  mju
 * Rename members to indicate pertain to current section.
 *
 * Revision 1.3  2012/04/05 15:57:52  mju
 * Add convertCoordTgtToSrc.
 *
 * Revision 1.2  2012/04/05 13:09:33  mju
 * Add methods to support using own work buffer.
 *
 * Revision 1.1  2012/04/02 17:18:01  mju
 * Initial revision
 *
 * \endif
**/

#ifndef	INC_SPATMOD_IMAGEHELPERRESAMPLE_H
#define	INC_SPATMOD_IMAGEHELPERRESAMPLE_H

#ifndef	INC_SPATMOD_IMAGEBUFFER_H
#include <spatmod/imagebuffer.h>
#endif

#ifndef	INC_SPATMOD_IMAGERESAMPLEMETHOD_H
#include <spatmod/imageresamplemethod.h>
#endif

#ifndef	INC_MI32_DOUBLEAR_H
#include <mi32/doublear.h>
#endif

#ifndef	INC_MI32_SIMPLEAR_H
#include <mi32/simplear.h>
#endif


#ifndef	GENERATING_DOXYGEN_OUTPUT
class TRANS2D_MAPGEN;
namespace SPATMOD {
	namespace IMAGE {
		class SAMPLEITERATOR;
		class SAMPLEITERATOR_CONST;
		class SECTION;
		class STAGE;
		}
	}
#endif


namespace SPATMOD {
namespace IMAGE {

//! Helper for resampling operations.
//! Provides method for processing single row.  Optimized to avoid memory allocation during row processing.
class HELPER_RESAMPLE {
	public:

		//! Flags for setup.
		enum FLAGS {
			FLAG_None =						0,
			FLAG_PullSinglePixels =		0x01,		//!< Pull pixels one at a time, for cases of large space between source pixels
			FLAG_CheckTargetValidity =	0x02,		//!< Use initial target validity values to determine if should process pixel
			};

		//! Constructor.
		HELPER_RESAMPLE ();

		//! Destructor.
		~HELPER_RESAMPLE ();

		//! Begin processing section using own work buffer.
		//! Work buffer will be resized if necessary.
		ERRVALUE BeginSection (
			const SECTION& SrcSection						//!< Section of source to resample from
			);

		//! Begin processing section using shared work buffer.
		ERRVALUE BeginSection (
			const SECTION& SrcSection,						//!< Section of source to resample from
			BUFFER& WorkBuffer								//!< Work buffer, must remain in existence during processing
			);

		//! Convert coordinates from target to source.
		//! Only points with a corresponding nonzero entry in CoordValid may be processed.  Any points that cannot
		//! be converted will have the corresponding CoordValid entry set to 0.
		ERRVALUE ConvertCoordTgtToSrc (
			const TRANS2D_MAPGEN& TransTgtToSrc,		//!< Transformation from target to source coordinates
			const DOUBLE_ARRAY<DPOINT2D>& TgtCoord,	//!< Target coordinates
			DOUBLE_ARRAY<DPOINT2D>& SrcCoord,			//!< Source coordinates returned, cannot be same array as TgtCoord
			UINT8 *CoordValid									//!< Source coordinate validity passed/returned, cannot be NULL
			) const;

		//! Free work buffer.
		void FreeWorkBuffer ()
			{ m_SrcBuffer.Destroy(); }

		//! Process row of pixels into target buffer.
		ERRVALUE ProcessRow (
			INT32 TgtRow,										//!< Row in target buffer
			BUFFER& TgtBuffer,								//!< Target buffer
			const DOUBLE_ARRAY<DPOINT2D>& SrcCoord		//!< Coordinates in source image, number of items determines number of output pixels
			);

		//! Reserve work buffer with specified initial maximum size.
		ERRVALUE ReserveWorkBuffer (
			INT32 MaxColumns,									//!< Maximum expected columns
			INT32 MaxRows										//!< Maximum expected rows
			);

		//! Setup to process specified source.
		ERRVALUE Setup (
			STAGE *pStageSrc,									//!< Pointer to source stage
			RESAMPLEMETHOD ResampleMethod,				//!< Resampling method
			FLAGS flags = FLAG_None							//!< Optional flags
			);

	private:
		#ifndef	GENERATING_DOXYGEN_OUTPUT
		BUFFER m_SrcBuffer;
		STAGE *m_pStageSrc;
		double m_bc1;
		double m_bc2;
		double m_bc3;
		double m_bc4;
		double m_bc5;
		int m_ksize;
		INT32 m_SectionCols;
		INT32 m_SectionCols1;
		INT32 m_SectionCols2;
		INT32 m_SectionRows;
		INT32 m_SectionRows1;
		INT32 m_SectionRows2;
		INT32 m_SectionBaseColumn;
		INT32 m_SectionBaseRow;
		UINT16 m_NumSamples;
		RESAMPLEMETHOD m_ResampleMethod;
		bool m_PullSinglePixels;
		bool m_CheckTgtValidity;
		bool m_IsAllUINT8;
		SIMPLE_ARRAY<const UINT8*> m_PtrSrcByte[4];
		SIMPLE_ARRAY<UINT8*> m_PtrTgtByte;
		SAMPLEITERATOR_CONST *m_itSrc[4];
		SAMPLEITERATOR *m_itTgt;

		void Free ();

		HELPER_RESAMPLE (const HELPER_RESAMPLE&);
		HELPER_RESAMPLE& operator= (const HELPER_RESAMPLE&);
		#endif
	};

#ifndef	GENERATING_DOXYGEN_OUTPUT
DEFINE_ENUM_OP_BITWISE(HELPER_RESAMPLE::FLAGS)
#endif

}	// End namespace IMAGE
}	// End namespace SPATMOD

#endif	// INC_SPATMOD_IMAGEHELPERRESAMPLE_H
